/*
 * Table StatisticChangesInApprovalProcess
 * En el momento de ejecutar la consulta, número de cambios registrados en el sistema que están en proceso de aprobación
 */

CREATE OR REPLACE TRIGGER tgr_StatisticChangesInApprovalProcess
AFTER INSERT OR UPDATE OF approvalStatusID OR DELETE ON Change
FOR EACH ROW
DECLARE
  v_inProcess NUMBER(1);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticChangesInApprovalProcess');

  IF inserting THEN
    v_inProcess := +1;
  ELSIF deleting AND :old.approvalStatusID = 0 THEN
    v_inProcess := -1;
  ELSIF (:old.approvalStatusID = 0) AND (:old.approvalStatusID != :new.approvalStatusID) THEN
    -- aprovalStatusID fue modificado
    v_inProcess := -1;
  ELSE
    v_inProcess := 0;
  END IF;

  IF v_inProcess != 0 THEN
    -- Actualiza la tabla StatisticChangesInApprovalProcess
    UPDATE StatisticChangesInApprovalProcess
    SET numChangesInApprovalProcess = numChangesInApprovalProcess + v_inProcess;
  END IF;

END;


/*
 * Table StatisticNotSucceedExecutionsLastYear
 *
 * En el último año, número de cambios aprobados cuya ejecución no haya sido correcta.
 * Se computa sobre la fecha de cambio, no la fecha de aprobación.
 * Se usa una tabla con un único registro, sobre la que se actualiza dos contadores:
 * último año (se entiende, ejercicio vencido anterior al actual)
 * y año en curso, de modo que, con el cambio de año, no sea necesario ejecutar COUNT
 * sobre los cambios del año que finaliza, que pasa a ser el último año
 */

CREATE OR REPLACE TRIGGER tgr_StatisticChangesNotSucceedExecutions
AFTER UPDATE OF executionResultID ON Change
FOR EACH ROW
DECLARE 
  v_currentYear CHAR(4) := EXTRACT(YEAR FROM SYSDATE);
  v_currentNotSucceedYear CHAR(4);
  v_changeYear CHAR(4);
  v_notSucceed CHAR(1);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticChangesNotSucceedExecutions');

  --si la ejecución no ha sido satisfactoria, requerirá un plan de acción
  SELECT actionPlan INTO v_notSucceed FROM executionResult WHERE executionResultID = :new.executionResultID;
  
  IF (v_notSucceed = 'Y') THEN
    SELECT currentYear INTO v_currentNotSucceedYear FROM StatisticNotSucceedExecutionsLastYear;
    
    IF v_currentYear > v_currentNotSucceedYear THEN
      UPDATE StatisticNotSucceedExecutionsLastYear 
      SET currentYear = v_currentYear, 
          numNotSucceedExecutionsPreviousYear = CASE WHEN TO_NUMBER(v_currentYear) = TO_NUMBER(v_currentNotSucceedYear) + 1 THEN numNotSucceedExecutionsCurrentYear ELSE 0 END,
          numNotSucceedExecutionsCurrentYear = 0;
    END IF;
    
    v_ChangeYear := EXTRACT(YEAR FROM :new.changeDate);

    IF v_changeYear = v_currentYear THEN
      UPDATE StatisticNotSucceedExecutionsLastYear
      SET numNotSucceedExecutionsCurrentYear = numNotSucceedExecutionsCurrentYear + 1;

    ELSIF v_ChangeYear = TO_CHAR(TO_NUMBER(v_currentYear) - 1) THEN
      UPDATE StatisticNotSucceedExecutionsLastYear
      SET numNotSucceedExecutionsPreviousYear = numNotSucceedExecutionsPreviousYear + 1;

    END IF;
  END IF;
END;


/*
 * Table StatisticTopTechnicalManagerCorrectExecutions
 *
 * Teniendo en cuenta toda la información existente en la BD, responsable técnico con más cambios ejecutados de manera correcta
 * Se mantiene el número de cambios ejecutados de manera correcta por cada responsable técnico
 * y en una consulta debería devolver la primera fila, ordenada por número de cambios, de mayor a menor
 * Para usar una sola fila, se tendría que actualizar con el responsable técnico y
 * el valor MAX de COUNT cuando technicalManagerID = :new.technicalManagerID AND executionResultID = 1
 */

CREATE OR REPLACE TRIGGER tgr_StatisticTopTechnicalManagerCorrectExecutions
AFTER UPDATE OF executionResultID ON Change
FOR EACH ROW
DECLARE 
  v_notSucceed CHAR(1);
  v_technicalManagerID NUMBER(4);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticTopTechnicalManagerCorrectExecutions');

  --si la ejecución ha sido satisfactoria, no requerirá un plan de acción
  --más acotadamente, podría preguntase por :new.executionResultID = 1
  SELECT actionPlan INTO v_notSucceed FROM executionResult WHERE executionResultID = :new.executionResultID;

  IF (v_notSucceed = 'N') THEN
    SELECT CASE WHEN substituteID IS NULL THEN usualID ELSE substituteID END INTO v_technicalManagerID
    FROM ChangeApprover
    WHERE changeID = :new.changeID AND rolID = 1;

    MERGE INTO StatisticTopTechnicalManagerCorrectExecutions a
    USING dual
    ON (technicalManagerID = v_technicalManagerID)

    WHEN MATCHED THEN
      UPDATE SET numCorrectExecutions = numCorrectExecutions + 1

    WHEN NOT MATCHED THEN
      INSERT (technicalManagerID, numCorrectExecutions)
      VALUES (v_technicalManagerID, 1);
  END IF;
END;


/*
 * Table StatisticClosedActionPlanLinesOnTimePercentageYear
 * 
 * Dado un año concreto, porcentaje de acciones definidas para ejecuciones no correctas cerradas 
 * en el tiempo inicialmente definido en el plan de acción correspondiente
 * El año se obtiene de la fecha de inicio del plan de acción (YEAR(opened))
 * Se evalúa en el momento de cierre de una línea del plan de acción
 * comparando su fecha de cierre con la fecha prevista de cierre (closing)
 * La consulta deberá devolver onTimeActions / totalActions * 100
 */

CREATE OR REPLACE TRIGGER tgr_StatisticClosedActionPlanLinesOnTimePercentageYear
AFTER UPDATE OF closed ON ChangeActionPlanLine
FOR EACH ROW
DECLARE 
  v_opened DATE;
  v_closing DATE;
  v_year CHAR(4);
  v_onTimeActionPlanLines NUMBER(1);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticClosedActionPlanLinesOnTimePercentageYear');

  SELECT opened, closing INTO v_opened, v_closing FROM ChangeActionPlan WHERE actionPlanID = :new.actionPlanID;

  IF TRUNC(:new.closed) > TRUNC(v_closing) THEN
    v_onTimeActionPlanLines := 1;
  ELSE
    v_onTimeActionPlanLines := 0;
  END IF;
  
  MERGE INTO StatisticClosedActionPlanLinesOnTimePercentageYear
  USING dual
  ON (year = v_year)

  WHEN MATCHED THEN
    UPDATE SET totalActionPlanLines = totalActionPlanLines + 1, onTimeActionPlanLines = onTimeActionPlanLines + v_onTimeActionPlanLines

  WHEN NOT MATCHED THEN
    INSERT (year, totalActionPlanLines, onTimeActionPlanLines)
    VALUES (v_year, 1, v_onTimeActionPlanLines);

END;


/*
 * Table StatisticGCABApprovalsCurrentYear
 * En el año en curso, número total de cambios aprobados en la GCAB
 * A diferencia de una estadística anterior, de último año, esta
 * se aplica al año en curso, obtenido de la fecha del sistema,
 * así que serán suficientes dos campos: año y número de cambios aprobados
 * Para contabilizar el número de cambios, la estadística se basará en
 * los cambios aprobados por el gestor de cambios, que traslada las votaciones del GCAB
 */

CREATE OR REPLACE TRIGGER tgr_StatisticGCABApprovalsCurrentYear
AFTER UPDATE OF approved ON ChangeApprover
FOR EACH ROW
DECLARE 
  v_currentYear CHAR(4) := EXTRACT(YEAR FROM SYSDATE);
  v_year CHAR(4);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticGCABApprovalsCurrentYear');

  IF (:new.rolID = 3 AND :new.approved = 'Y') THEN
    -- cuando se registra el voto del gestor de cambios, si es positivo
    -- se incrementará el número de cambios aprovados en la GCAB
    SELECT currentYear INTO v_year FROM StatisticGCABApprovalsCurrentYear;

    IF v_year IS NULL THEN
      INSERT INTO StatisticGCABApprovalsCurrentYear
      (currentYear, numGCABApprovalsCurrentYear) VALUES
      (v_currentYear, 1);

    ELSIF v_year != v_currentYear THEN
      UPDATE StatisticGCABApprovalsCurrentYear SET
      currentYear = v_currentYear, numGCABApprovalsCurrentYear = 1;

    ELSE
      UPDATE StatisticGCABApprovalsCurrentYear SET
      numGCABApprovalsCurrentYear = numGCABApprovalsCurrentYear + 1;
    END IF;
  END IF;
END;


/*
 * Table StatisticDisapprovedChangesYear
 * En un año concreto, número de cambios que no fueron aprobados y, consecuentemente, no se ejecutaron
 */

CREATE OR REPLACE TRIGGER tgr_StatisticDisapprovedChangesYear
AFTER UPDATE OF approvalStatusID ON Change
FOR EACH ROW
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticDisapprovedChangesYear');

  IF (:new.approvalStatusID > 1)  THEN
    -- se asignó Rechazado (ID=2) a approvalStatusID

    MERGE INTO StatisticDisapprovedChangesYear
    USING dual
    ON (year = EXTRACT(YEAR FROM :new.changeDate))

    WHEN MATCHED THEN
      UPDATE SET numDisapprovedChanges = numDisapprovedChanges + 1

    WHEN NOT MATCHED THEN
      INSERT (year, numDisapprovedChanges)
      VALUES (EXTRACT(YEAR FROM :new.changeDate), 1);

  END IF;
END;


/*
 * Table StatisticReplannedChangesPercentageRegion
 * Dada una región geográfica concreta, porcentaje de cambios que su ejecución se replanificó (ID 3)
 */

CREATE OR REPLACE TRIGGER tgr_StatisticReplannedChangesPercentageRegion
AFTER UPDATE OF executionResultID ON Change
FOR EACH ROW
DECLARE 
  v_replanned CHAR(1);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticReplannedChangesPercentageRegion');

  IF :new.executionResultID = 3 THEN
    v_replanned := 1;
  ELSE
    v_replanned := 0;
  END IF;

  MERGE INTO StatisticReplannedChangesPercentageRegion a
  USING (SELECT geographicZoneID FROM ChangeGeographicScope WHERE changeID = :new.ChangeID) b
  ON (a.geographicZoneID = b.geographicZoneID)

  WHEN MATCHED THEN
    UPDATE SET totalExecution = totalExecution + 1, replannedExecution = replannedExecution + v_replanned

  WHEN NOT MATCHED THEN
    INSERT (geographicZoneID, totalExecution, replannedExecution)
    VALUES (b.geographicZoneID, 1, v_replanned);

END;


/*
 * Table StatisticNonCompliancesDetectedLastSixMonths
 *
 * En los últimos 6 meses, número total de incumplimientos detectados durante las auditorías realizadas
 *
 */

CREATE OR REPLACE TRIGGER tgr_StatisticNonCompliancesSixMonths
AFTER UPDATE OF nonCompliance ON ChangeAuditDetail
FOR EACH ROW
DECLARE 
  v_currentLastMonth DATE;
  v_tableLastMonth DATE;
  v_changeDate DATE;
  v_monthsTable NUMBER(4);
  v_monthsChange NUMBER(4);
 
  TYPE monthArrayType IS VARRAY(6) OF NUMBER(2);
  monthArray monthArrayType;
  --monthArray DBMS_SQL.NUMBER_TABLE;
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticNonCompliancesSixMonths');

  IF (:new.nonCompliance = 'Y') THEN
    v_currentLastMonth := TRUNC(ADD_MONTHS(SYSDATE, -1), 'MM');  --primer día del último mes vencido

    SELECT TO_DATE(lastMonth, 'YYYYMM'), month1, month2, month3, month4, month5, month6 
    INTO v_tableLastMonth, monthArray(1), monthArray(2), monthArray(3), monthArray(4), monthArray(5), monthArray(6)
    FROM StatisticNonCompliancesSixMonths;  --primer día del mes de la última estadística

    SELECT MONTHS_BETWEEN(v_currentLastMonth, v_tableLastMonth) INTO v_monthsTable FROM DUAL;

    IF v_monthsTable < 6 THEN
      FOR i IN REVERSE monthArray.First + v_monthsTable..monthArray.Last LOOP
        monthArray(i) := monthArray(i - v_monthsTable);
      END LOOP;

      FOR i IN monthArray.First..v_monthsTable LOOP
        monthArray(i) := 0;
      END LOOP;
    END IF;

    SELECT changeDate INTO v_changeDate FROM change WHERE changeID = (SELECT changeID FROM ChangeAudit WHERE auditID = :new.auditID);
    
    SELECT MONTHS_BETWEEN(v_currentLastMonth, v_changeDate) INTO v_monthsChange FROM DUAL;
    
    IF v_monthsChange < 6 THEN
      monthArray(v_monthsChange + 1) := monthArray(v_monthsChange + 1) + 1;
    END IF;
    
    IF (v_monthsTable < 6) OR (v_monthsChange < 6) THEN
      UPDATE StatisticNonCompliancesSixMonths SET
      lastMonth = TO_CHAR(v_currentLastMonth, 'YYYYMM'),
      month1 = monthArray(1),
      month2 = monthArray(2),
      month3 = monthArray(3),
      month4 = monthArray(4),
      month5 = monthArray(5),
      month6 = monthArray(6);
    END IF;
  END IF;
END;


/*
 * Table StatisticOpenedActionPlans
 *
 * En el momento de ejecutar la consulta, número total de planes de acción sobre ejecuciones no correctas que están abiertos
 */

CREATE OR REPLACE TRIGGER tgr_StatisticOpenedActionPlans
AFTER INSERT OR UPDATE OF closed OR DELETE ON ChangeActionPlan
FOR EACH ROW
DECLARE
  v_opened INTEGER;
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticOpenedActionPlans');

  IF inserting THEN
    v_opened := +1;
  ELSIF deleting AND :old.closed IS NOT NULL THEN
    v_opened := -1;
  ELSIF (:old.closed IS NULL) AND (:old.closed != :new.closed) THEN
    -- aprovalStatusID fue modificado
    v_opened := -1;
  ELSE
    v_opened := 0;
  END IF;

  IF v_opened != 0 THEN
    -- Actualiza la tabla StatisticOpenedActionPlans
    UPDATE StatisticOpenedActionPlans
    SET numOpenedActionPlans = numOpenedActionPlans + v_opened;
  END IF;
END;


/*
 * Table StatisticTopTechnicalManagerChangesInProgress
 * En un momento dado, responsable técnico con más cambios en curso
 */

CREATE OR REPLACE TRIGGER tgr_StatisticTopTechnicalManagerChangesInProgress
AFTER INSERT OR UPDATE ON ChangeApprover
FOR EACH ROW
DECLARE
  v_technicalManagerID NUMBER(4);
BEGIN
  IF (:new.rolID = 1) THEN
    --DBMS_OUTPUT.PUT_LINE('tgr_StatisticTopTechnicalManagerChangesInProgress');

    IF (:old.approved != :new.approved) THEN
      --Votó, restamos de la estadística de cambios en progreso
      UPDATE StatisticTopTechnicalManagerChangesInProgress
      SET numChangesInProgress = numChangesInProgress - 1
      WHERE technicalManagerID = CASE WHEN :new.substituteID IS NOT NULL THEN :new.substituteID ELSE :new.usualID END; 

    ELSIF inserting OR (:old.substituteID != :new.substituteID) THEN
      --Alta del responsable técnico o sustitución
      IF (:old.substituteID != :new.substituteID) THEN
        --Restar al responsable técnico inicial
        v_technicalManagerID := :new.substituteID;

        UPDATE StatisticTopTechnicalManagerChangesInProgress
        SET numChangesInProgress = numChangesInProgress - 1
        WHERE technicalManagerID = :new.usualID; 
      ELSE
        v_technicalManagerID := :new.usualID;
      END IF;
        
      MERGE INTO StatisticTopTechnicalManagerChangesInProgress a
      USING dual
      ON (a.technicalManagerID = v_technicalManagerID)

      WHEN MATCHED THEN
        UPDATE SET numChangesInProgress = numChangesInProgress + 1

      WHEN NOT MATCHED THEN
        INSERT (technicalManagerID, numChangesInProgress)
        VALUES (v_technicalManagerID, 1);
    END IF;
  END IF;
END;

/*
CREATE OR REPLACE TRIGGER tgr_StatisticTopTechnicalManagerChangesInProgress2
AFTER UPDATE OF approvalStatusID ON Change
FOR EACH ROW
DECLARE
  v_technicalManagerID NUMBER(4);
BEGIN
  DBMS_OUTPUT.PUT_LINE('tgr_StatisticTopTechnicalManagerChangesInProgress2');

  IF (:old.approvalStatusID = 0) THEN --AND (:old.approvalStatusID != :new.approvalStatusID) THEN
    SELECT CASE WHEN substituteID IS NULL THEN usualID ELSE substituteID END INTO v_technicalManagerID
    FROM ChangeApprover
    WHERE changeID = :old.changeID AND rolID = 1;

    UPDATE StatisticTopTechnicalManagerChangesInProgress
    SET numChangesInProgress = numChangesInProgress - 1
    WHERE technicalManagerID = v_technicalManagerID; 
  END IF;

END;


CREATE OR REPLACE TRIGGER tgr_StatisticTopTechnicalManagerChangesInProgress2
AFTER UPDATE OF approved ON ChangeApprover
FOR EACH ROW
WHEN (:new.rolID = 1)
DECLARE
  v_changeApproved CHAR(1);
  v_technicalManagerID NUMBER(4);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('trg_updateApprover');

  SELECT
    CASE WHEN EXISTS(SELECT * FROM ChangeApprover WHERE ChangeID = :new.changeID AND approved IS NULL) THEN NULL
         WHEN EXISTS(SELECT * FROM ChangeApprover WHERE ChangeID = :new.changeID AND approved = 'N') THEN 'N'
         ELSE 'Y' END into v_changeApproved
  FROM DUAL;

  IF v_changeApproved IS NOT NULL THEN
    IF :new.substituID IS NOT NULL
      v_technicalManagerID := :new.substituteID
    ELSE
      v_technicalManagerID := :new.usualID;

    UPDATE StatisticTopTechnicalManagerChangesInProgress
    SET numChangesInProgress = numChangesInProgress - 1
    WHERE technicalManagerID = v_technicalManagerID; 

  END IF;
END;
*/

/*
 * Table StatisticTopTechnicalManagerChangesInProgress
 * Teniendo en cuenta todos los datos de la BD, persona concreta con rol de aprobador que ha estado sustituida más veces por indisponibilidad no planificada
 */

CREATE OR REPLACE TRIGGER tgr_StatisticSubstitutionApprover
AFTER INSERT OR UPDATE ON ChangeApprover
FOR EACH ROW
DECLARE
  v_oldPlanned CHAR(1);
  v_newPlanned CHAR(1);
  v_increment NUMBER(1);
  v_personID NUMBER(4);
BEGIN
  --DBMS_OUTPUT.PUT_LINE('tgr_StatisticSubstitutionApprover');

  IF (:old.substitutionReasonID != :new.substitutionReasonID) THEN
    SELECT planned INTO v_oldPlanned FROM SubstitutionReason WHERE substitutionReasonID = :old.substitutionReasonID;
    SELECT planned INTO v_newPlanned FROM SubstitutionReason WHERE substitutionReasonID = :new.substitutionReasonID;

    IF (v_oldPlanned = 'Y') OR (v_newPlanned = 'Y') THEN  -- ¿la causa de sustitución es no planificada?

      IF (deleting AND (v_oldPlanned = 'Y')) OR (v_oldPlanned = 'Y') AND (v_newPlanned != 'Y') THEN
        v_increment := -1;
      ELSIF (v_oldPlanned != 'Y') AND (v_newPlanned = 'Y') THEN
        v_increment := +1;
      ELSE
        v_increment := 0;
      END IF;

      IF v_increment != 0 THEN
        MERGE INTO StatisticSubstitutionApprover a
        USING dual
        ON (a.personID = :new.usualID)
        WHEN MATCHED THEN
          UPDATE SET numSubstitutions = numSubstitutions + v_increment
        WHEN NOT MATCHED THEN
          INSERT (personID, numSubstitutions)
          VALUES (:new.usualID, v_increment);
      END IF;

    END IF;
  END IF;
END;

